#include <stdlib.h>
#include <stdio.h>
#include <conio.h>
#include <graphics.h>
#include <string.h>
#include <math.h>
#include <alloc.h>
#include <limits.h>

typedef unsigned int uint;
typedef unsigned long ulong;

#define V_MAX 2000
#define L_MAX 2000
#define SEC_MAX 400
#define SEG_MAX 2800
#define SS_MAX 1000
#define N_MAX 900
#define LL_MAX 16384
#define VL_MAX 2048

void generate_nodes(void);
void add_node_list(int far *list, int *index, int line, int v1, int v2);
uint nodize(uint size, int far *list);
int resolve_space(int line, int size, int far *list,
	int far *left_list, int far *right_list, int *lx1, int *ly1,
	int *lx2, int *ly2, int *rx1, int *ry1, int *rx2, int *ry2, int test);
void add_node_list2(int far *list, int index, int line, int v1, int v2,
	int *x1, int *y1, int *x2, int *y2);
int make_seg(int line, int v1, int v2);
int calc_line_cross(int *y, int x1, int y1, int v2, int v3, int v4);
uint calc_angle(int x1, int y1, int x2, int y2);
int adjusted_angle(uint angle, int x1, int y1, int x2, int y2);
void adjust_limit(int x, int y, int *xmin, int *ymin, int *xmax, int *ymax);
void error(char *msg);
void fatal_error(char *msg);

/* these are global variables because I don't need recursion building new
	 sets of all these variables
*/

int optimize; /* method of optimizing node generation */
int v_size_backup;
int seg_orig;
int i, j; /* general loop indexing variables */
int nodeline; /* current nodeline */
int convex; /* set if polygon is convex */
int err; /* error code */
uint lcount; /* # of lines to the left of nodeline */
uint rcount; /* # of lines to the right of nodeline */
int bal; /* difference between lcount and rcount (0 = perfect balance) */
int splits; /* # of lines nodeline splits */
int left_size; /* lcount of best case */
int right_size; /* rcount of best case */
int balance; /* best case balance */
int splits_min; /* best case of minimum splits */
uint link2; /* node/ssec left side link */

uint v_size; /* # of vertexes */
uint max_vertex; /* # of vertexes used by lines (below seg used ones) */
uint l_size; /* # of linedefs */
uint s_size; /* # of sidedefs */
uint n_size; /* size of nodes block */
uint ss_size; /* # of special sectors */
uint sec_size; /* # of sectors */
uint seg_size; /* # of segments */

struct {
	int v1;
	int v2;
	uint attrib;
	int type;
	int trig;
	int sd1;
	int sd2;
} far linedefs[L_MAX];

struct {
	int x;
	int y;
} far vertexes[V_MAX];

struct seg_struct {
	int v1;
	int v2;
	uint angle;
	int linedef;
	int dir;
	int dist;
} far segs[SEG_MAX];

struct ss_struct {
	int size;
	int segptr;
} far ssectors[SS_MAX];

struct n_struct {
	int x;
	int y;
	int dx;
	int dy;
	struct {
		int y1;
		int y2;
		int x1;
		int x2;
	} node1;
	struct {
		int y1;
		int y2;
		int x1;
		int x2;
	} node2;
	uint link1;
	uint link2;
} far nodes[N_MAX];

void generate_nodes(void)
{
	int index, v1, v2, far *list;
	int i;
	long j;
	struct ss_struct ssec;

	optimize = 2;
	n_size = ss_size = seg_size = 0;
	v_size = max_vertex;
	if ((list = farmalloc(s_size * 6)) != NULL)
	{
		index = err = 0;
		for (i=0; i<l_size; i++)
		{
			v1 = linedefs[i].v1;
			v2 = linedefs[i].v2;

			if (linedefs[i].sd1 != -1)
				add_node_list(list, &index, i, v1, v2);
			if (linedefs[i].sd2 != -1)
				add_node_list(list, &index, i, v2, v1);
		}

		if (nodize(index/3, list) != 65535)
		{
			farfree(list);
			return;
		}
		farfree(list);
	}
	if (!err)
		error("Out of far memory!");
	if (err == 1)
		error("Too many Nodes!");
	if (err == 2)
		error("Too many Segments!");
	if (err == 3)
		error("Too many Sub Sectors!");
	if (err == 4)
		error("Too many Vertexes!");
	n_size = ss_size = seg_size = 0;
	v_size = max_vertex;
	return;
}

/* used by generate_nodes() to built a list of starting segs */

void add_node_list(int far *list, int *index, int line, int v1, int v2)
{
	list[(*index)++] = line;
	list[(*index)++] = v1;
	list[(*index)++] = v2;
	return;
}

/*  recursive node generator  */

uint nodize(uint size, int far *list)
{
	int x1, y1, x2, y2, lx1, ly1, lx2, ly2, rx1, ry1, rx2, ry2;
	int num, far *left_list, far *right_list;
	uint link1;

	convex = 1;
	balance = splits_min = size;
	nodeline = -1;
	for (i=0, j=1; i<size; i++, j +=3)
	{
		x1 = vertexes[list[j]].x;
		x2 = vertexes[list[j+1]].x;
		y1 = vertexes[list[j]].y;
		y2 = vertexes[list[j+1]].y;

		if (resolve_space(i, size, list, left_list, right_list,
			&lx1, &ly1, &lx2, &ly2, &rx1, &ry1, &rx2, &ry2, 1))
				return 65535; /* test nodeline to see what we would get */

		if (lcount && rcount) /* lines on both sides of nodelist */
			convex = 0; /* if ever so, can't be convex polygon */

		bal = abs(lcount - rcount);
		num = 0;
		switch (optimize)
		{
			case 0: /* optimize for balance */
				if (bal < balance)
					num++;
				if (bal == balance && splits < splits_min)
					num++;
				break;
			case 1: /* optimize for least splits */
				if (splits < splits_min)
					num++;
				if (splits == splits_min && bal < balance)
					num++;
				break;
			case 2: /* optimize between both */
				if ((bal + splits) <= (balance + splits_min))
					if (bal < balance)
						num++;
				break;
		}
		if (!lcount || !rcount) /* all lines on one side */
			num = 0; /* ineligible, because we can't split anymore */

		if (num)
		{
			balance = bal;
			splits_min = splits;
			left_size = lcount;
			right_size = rcount;
			nodeline = i; /* best case for nodeline so far */
		}
	}

	if (convex) /* is a convex polygon, so generate an ssec */
	{
		if (ss_size == SS_MAX)
		{
			err = 3;
			return 65535;
		}
		ssectors[ss_size].segptr = seg_size;
		ssectors[ss_size++].size = size;

		for (i=0; i<size; i++)
		{
			num = i * 3;
			if (make_seg(list[num], list[num+1], list[num+2]))
			{
				err = 2;
				return 65535;
			}
		}
		return (32767 + ss_size); /* ssector link value */
	}
	if (nodeline == -1)
		fatal_error("No nodeline selected");

	num = nodeline * 3;
	x1 = vertexes[list[num+1]].x;
	x2 = vertexes[list[num+2]].x;
	y1 = vertexes[list[num+1]].y;
	y2 = vertexes[list[num+2]].y;
	if (left_size)
		if ((left_list = farmalloc(left_size * 6)) == NULL)
			return 65535;
	if (right_size)
		if ((right_list = farmalloc(right_size * 6)) == NULL)
		{
			farfree(left_list);
			return 65535;
		}

	if (resolve_space(nodeline, size, list, left_list, right_list,
		&lx1, &ly1, &lx2, &ly2, &rx1, &ry1, &rx2, &ry2, 0)) /* get for real */
	{
		farfree(left_list);
		farfree(right_list);
		return 65535;
	}
	if (!lcount || !rcount)
		fatal_error("convex polygon not detected");

	num = left_size; /* global var we need, and nodize recursion will scramble */
	link1 = nodize(right_size, right_list);
	farfree(right_list); /* don't need it anymore */
	if (link1 < 65535)
	{
		link2 = nodize(num, left_list);
		if (link2 < 65535)
		{
			if (n_size < N_MAX)
			{
				farfree(left_list);
				nodes[n_size].x = x1;
				nodes[n_size].y = y1;
				nodes[n_size].dx = x2 - x1;
				nodes[n_size].dy = y2 - y1;
				nodes[n_size].node1.x1 = rx1;
				nodes[n_size].node1.y1 = ry1;
				nodes[n_size].node1.x2 = rx2;
				nodes[n_size].node1.y2 = ry2;
				nodes[n_size].node2.x1 = lx1;
				nodes[n_size].node2.y1 = ly1;
				nodes[n_size].node2.x2 = lx2;
				nodes[n_size].node2.y2 = ly2;
				nodes[n_size].link1 = link1;
				nodes[n_size].link2 = link2;
				return n_size++; /* link to this node */
			}
			err = 1;
		}
	}
	farfree(left_list);
	return 65535; /* error, return back through recursions */
}

/*  make left and right lists from master list  */

int resolve_space(int line, int size, int far *list,
	int far *left_list, int far *right_list, int *lx1, int *ly1,
	int *lx2, int *ly2, int *rx1, int *ry1, int *rx2, int *ry2, int test)
{
	int i, v1, v2, v3, v4, nv1, nv2, x, y, xx, yy, num;
	int angle1, angle2, left, right, line2;
	uint angle;

	lcount = rcount = splits = 0; /* reset counts */
	*lx1 = *rx1 = INT_MAX; /* reset limits */
	*lx2 = *rx2 = INT_MIN;
	*ly1 = *ry1 = INT_MAX;
	*ly2 = *ry2 = INT_MIN;

	num = line * 3;
	nv1 = list[num + 1]; /* nodeline vertexes */
	nv2 = list[num + 2];
	x = vertexes[nv1].x;
	y = vertexes[nv1].y;
	angle = calc_angle(x, y, vertexes[nv2].x, vertexes[nv2].y);

	for (i=0; i<size; i++)
	{
		left = right = 0;
		num = i * 3;
		line2 = list[num];
		v1 = v3 = list[num + 1];
		v2 = v4 = list[num + 2];

		if (v1 == nv1)
			angle1 = 0;
		else {
			angle1 = adjusted_angle(angle, x, y, vertexes[v1].x, vertexes[v1].y);
			if (angle1 == INT_MIN) /* on line, but behind vertex 1 */
				angle1 = 0;
		}

		if (v2 == nv1)
			angle2 = 0;
		else {
			angle2 = adjusted_angle(angle, x, y, vertexes[v2].x, vertexes[v2].y);
			if (angle2 == INT_MIN)
				angle2 = 0;
		}

		if ((!angle1) && (!angle2)) /* on line */
		{
			if (adjusted_angle(angle, vertexes[v1].x, vertexes[v1].y,
				vertexes[v2].x, vertexes[v2].y))
				left = 1;
			else
				right = 1;

		} else if (angle1 >= 0 && angle2 >= 0) /* totally right of line */
			right = 1;

		else if (angle1 < 1 && angle2 < 1) /* totally left of line */
			left = 1;

		else { /* nodeline must split this line */
			left = right = 1; /* line on each side */
			splits++; /* inc split count */
		}

		if (!test) /* if not in test mode, build the lists */
		{
			if (right)
				add_node_list2(right_list, rcount++, line2, v1, v2,
					rx1, ry1, rx2, ry2);
			if (left)
				add_node_list2(left_list, lcount++, line2, v3, v4,
					lx1, ly1, lx2, ly2);
		} else {
			if (right) /* still need to track counts, though */
				rcount++;
			if (left)
				lcount++;
		}
	}
	return 0;
}

void add_node_list2(int far *list, int index, int line, int v1, int v2,
	int *x1, int *y1, int *x2, int *y2)
{
	uint idx;

	adjust_limit(vertexes[v1].x, vertexes[v1].y, x1, y1, x2, y2);
	adjust_limit(vertexes[v2].x, vertexes[v2].y, x1, y1, x2, y2);

	idx = index * 3;
	list[idx] = line;
	list[idx+1] = v1;
	list[idx+2] = v2;
	return;
}

int make_seg(int line, int v1, int v2)
{
	int v, vv, x, y, x1, y1, x2, y2, dir, dist;
	uint angle1, angle2;

	if (seg_size == SEG_MAX)
		return -1;

	v = linedefs[line].v1;
	vv = linedefs[line].v2;
	x = vertexes[v].x;
	y = vertexes[v].y;

	dir = dist = 0;
	angle1 = calc_angle(x1, y1, x2, y2);
	angle2 = calc_angle(x, y, vertexes[vv].x, vertexes[vv].y);
	if (labs((long int) angle2 - angle1) > 16384) /* opposite direction */
	{
		dir = 1;
		v = linedefs[line].v2;
		x = vertexes[v].x;
		y = vertexes[v].y;
	}
	if (v1 != v)
	{
		x = abs(x - x1);
		y = abs(y - y1);
		dist = sqrt((double) x * x + (double) y * y);
	}

	segs[seg_size].v1 = v1;
	segs[seg_size].v2 = v2;
	segs[seg_size].linedef = line;
	segs[seg_size].angle = angle1;
	segs[seg_size].dir = dir;
	segs[seg_size++].dist = dist;
	return 0;
}

/* calculate the (x,y) where 2 lines cross */

int calc_line_cross(int *y, int x1, int y1, int v2, int v3, int v4)
{
	int x2, y2, x3, y3, x4, y4;
	long dx1, dx2, dy1, dy2, yy, p1, p2, p3;

	x2 = vertexes[v2].x;
	x3 = vertexes[v3].x;
	x4 = vertexes[v4].x;
	y2 = vertexes[v2].y;
	y3 = vertexes[v3].y;
	y4 = vertexes[v4].y;

	dx1 = x1 - x2;
	dx2 = x3 - x4;
	dy1 = y1 - y2;
	dy2 = y3 - y4;

	if (!dx1) /* line 1 is vertical */
	{
		if (!dy2) /* line 2 horizontal? */
		{
			*y = y3;
			return x1;
		}
		*y = (float) dy2 * (x1 - x3) / dx2 + y3;
		return x1;
	}

	if (!dx2) /* line 2 is vertical */
	{
		if (!dy1) /* line 1 horizontal? */
		{
			*y = y1;
			return x3;
		}
		*y = (float) dy1 * (x3 - x1) / dx1 + y1;
		return x3;
	}

	if (!dy1) /* line 1 is horizontal */
	{
		*y = y1;
		if (!dx2) /* line 2 vertical? */
			return x3;
		return ((float) dx2 * (*y - y3) / dy2 + x3);
	}

	if (!dy2) /* line 2 is horizontal */
	{
		*y = y3;
		if (!dx1) /* line 1 vertical? */
			return x1;
		return ((float) dx1 * (*y - y1) / dy1 + x1);
	}

	p1 = dx1 * dy2;
	p2 = dx2 * dy1;
	p3 = dy1 * dy2;

	yy = (y1*p1 - y3*p2 + x3*p3 - x1*p3) / (p1 - p2);
	*y = yy;
	return (float) dx1 * (yy - y1) / dy1 + x1;
}

/*  calculate the angle of line (x1,y1)-(x2,y2).  (x1,y1) is origin  */

uint calc_angle(int x1, int y1, int x2, int y2)
{
	int x, y;
	uint ang;
	double angle;

	x = abs(x2 - x1);
	y = abs(y2 - y1);
	angle = M_PI_2;
	if (x != 0)
		angle = atan( (float) y/x);
	ang = angle * 32768 / M_PI;
	x = ((x2-x1) < 0)*2 + ((y2-y1) < 0);
	if (x == 1)
		ang = 65536 - ang;
	if (x == 2)
		ang = 32768 - ang;
	if (x == 3)
		ang += 32768;
	return ang;
}

/* return angle of line (x1,y1)-(x2,y2) using <angle> as the reference */

int adjusted_angle(uint angle, int x1, int y1, int x2, int y2)
{
	long angle2;

	angle2 = (long) angle - calc_angle(x1, y1, x2, y2);
	if (angle2 > INT_MAX)
		angle2 -= 65536;
	if (angle2 < INT_MIN)
		angle2 += 65536;
	return angle2;
}

void adjust_limit(int x, int y, int *xmin, int *ymin, int *xmax, int *ymax)
{
	if (x < *xmin)
		*xmin = x;
	if (x > *xmax)
		*xmax = x;
	if (y < *ymin)
		*ymin = y;
	if (y > *ymax)
		*ymax = y;
	return;
}

void error(char *msg)
{
	fprintf(stderr, "Error: %s\n", msg);
	getch();
	return;
}

void fatal_error(char *msg) /* terminate with error */
{
	fprintf(stderr, "Fatal error: %s\n", msg);
	exit(1);
}
